import logging
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from modern_pylogging import import_checker
from modern_pylogging.contextvars_helpers import get_log_extra

if import_checker.is_picologging_installed:
    import picologging


if TYPE_CHECKING:
    LogRecord: TypeAlias = logging.LogRecord | picologging.LogRecord  # type:ignore[possibly-undefined]


class ConsoleFormatterMixin:
    def format(self, record: 'LogRecord') -> str:
        extra_data = self._prepare_log_extra(record)
        # Build key=value string
        k_v_strings = (f'{k} = {v}' for k, v in extra_data.items())  # noqa: WPS111
        extra_data_str = ', '.join(k_v_strings)
        record.extra_data_str = f'{{{extra_data_str}}}' if extra_data_str else ''  # type:ignore[union-attr]
        return cast('str', super().format(record))  # type:ignore[misc]

    def _prepare_log_extra(self, record: 'LogRecord') -> dict[str, Any]:
        extra = get_log_extra()
        if extra:
            return extra
        if hasattr(record, 'binded_ctx_extra'):
            return cast('dict[str, Any]', record.binded_ctx_extra)
        return {}


class ConsoleFormatterLogging(ConsoleFormatterMixin, logging.Formatter):
    """Class to use with stdlib logging"""


if import_checker.is_picologging_installed:

    class ConsoleFormatterPicologging(ConsoleFormatterMixin, picologging.Formatter):
        """Class to use with picologging library"""



import datetime as dt
import logging
import socket
from collections.abc import Callable
from functools import lru_cache
from pathlib import Path
from traceback import format_tb
from typing import TYPE_CHECKING, Any, Final, Literal, TypeAlias, cast

from modern_pylogging import import_checker
from modern_pylogging.config_api import _get_default_json_dumps_module
from modern_pylogging.contextvars_helpers import get_log_extra
from modern_pylogging.helper_utils import get_env

if import_checker.is_picologging_installed:
    import picologging

if TYPE_CHECKING:
    LogRecord: TypeAlias = logging.LogRecord | picologging.LogRecord  # type:ignore[possibly-undefined]


# skip default LogRecord attributes
RESERVED_ATTRS: Final[set[str]] = {  # noqa: WPS407
    'args',
    'asctime',
    'created',
    'exc_info',
    'filename',
    'funcName',
    'levelname',
    'levelno',
    'lineno',
    'message',
    'module',
    'msecs',
    'msg',
    'name',
    'pathname',
    'process',
    'processName',
    'relativeCreated',
    'stack_info',
    'thread',
    'threadName',
    'taskName',
    'binded_ctx_extra',  # from QueueHandlerContextVarsHappyPy312
}


def grab_record_extra_fields(record: 'LogRecord', reserved: set[str]) -> dict[str, Any]:
    """Extracts extra attributes from LogRecord object."""
    extra_fields = {}
    for key, value in record.__dict__.items():  # noqa: WPS110
        if key in reserved:
            continue
        # this allows to have numeric keys
        if hasattr(key, 'startswith') and key.startswith('_'):
            continue
        extra_fields[key] = value
    return extra_fields


def provide_json_dumps_func(json_dumps_module: Literal['json', 'orjson']) -> Callable[[Any], str]:
    if json_dumps_module == 'orjson':
        from modern_pylogging.orjson_helper import json_dumps

        return json_dumps
    from modern_pylogging.json_helper import json_dumps

    return json_dumps


class JsonFormatterMixin:
    def __init__(
        self,
        *args: Any,
        json_dumps_module: Literal['json', 'orjson'] | None = None,
        capture_extra_fields: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        json_dumps_module = json_dumps_module or _get_default_json_dumps_module()
        self.json_dumps = provide_json_dumps_func(json_dumps_module)
        self.capture_extra_fields = capture_extra_fields

    def format(self, record: 'LogRecord') -> str:
        message = self._prepare_log_dict(record)
        return self.json_dumps(message)

    def _prepare_log_extra(self, record: 'LogRecord') -> dict[str, Any]:
        extra = get_log_extra()
        if extra:
            return extra
        if hasattr(record, 'binded_ctx_extra'):
            return cast('dict[str, Any]', record.binded_ctx_extra)
        return {}

    def _prepare_log_dict(self, record: 'LogRecord') -> dict[str, Any]:  # noqa: WPS210
        log_record: dict[str, Any] = {}
        log_record['timestamp'] = timestamp_to_iso(record.created)
        log_record['level'] = record.levelname
        log_record['message'] = record.getMessage()
        log_record['params'] = {
            'call_filepath': f'{record.pathname}:{record.lineno}',
            'logger_name': record.name,
        }
        if record.exc_info:
            exc_type, exc_value, tb = record.exc_info
            log_record['error'] = {
                'code': exc_type.__class__.__name__,
                'message': repr(exc_value),
                'stack': format_tb(tb),
                'params': {},
            }
        default_data = get_logging_defaults()
        extra_request_data = self._prepare_log_extra(record)
        update_fields_with_nested(log_record, default_data)
        update_fields_with_nested(log_record, extra_request_data)
        if self.capture_extra_fields:
            local_extra_fields = grab_record_extra_fields(record, reserved=RESERVED_ATTRS | set(default_data))
            update_fields_with_nested(log_record, local_extra_fields)
        return log_record


class JsonFormatterLogging(JsonFormatterMixin, logging.Formatter):
    """Formatter to use with stdin logging library"""


if import_checker.is_picologging_installed:

    class JsonFormatterPicologging(JsonFormatterMixin, picologging.Formatter):
        """Formatter for use with picologging library"""


def timestamp_to_iso(timestamp: float) -> str:
    return (
        dt.datetime.fromtimestamp(timestamp, tz=dt.timezone.utc)
        .isoformat(timespec='milliseconds')
        .replace('+00:00', 'Z')
    )


@lru_cache(maxsize=1)
def get_logging_defaults() -> dict[str, Any]:
    """Reads default env variables. This vars stays the same between restarts."""
    return {
        'env': get_env(['ENV', 'LOGGING_ENV'], 'dev'),
        'system': get_env(['LOGGING_SYSTEM', 'APP_NAME'], Path.cwd().name),
        'inst': get_env(['POD_NAME', 'LOGGING_INST'], socket.gethostname()),
    }


def update_fields_with_nested(
    source_data: dict[str, Any],
    updated_fields: dict[str, Any],
) -> dict[str, Any]:
    """
    Update a keys in a dictionary.
    If patch to the key contains dots it means nested dictionary.
    Creates new nested dictionaries if necessary.
    'a.b.c': 'hello' -> {'a': {'b': {'c': 'hello'}}}
    """
    for field_name, field_value in updated_fields.items():
        here = source_data
        keys = field_name.split('.')
        for key in keys[:-1]:
            here = here.setdefault(key, {})
        here[keys[-1]] = field_value
    return source_data



import os
from typing import Any, TypeVar

T = TypeVar('T')  # noqa: WPS111


def get_env(envs: list[str], default: T) -> str | T:
    """
    Tries to get a value from any env variable from the list.
    If there's no value, return default.
    :param envs: list of env variable names
    :param default: default value if all env variables is not set
    """
    env_value = None
    for env in envs:
        env_value = os.getenv(env, None)
        if env_value is not None:
            break
    if env_value is None:
        return default
    return env_value


def get_logging_level() -> str | int:
    level = get_env(['LOGGING_LEVEL'], 'INFO')
    if level.isdigit():
        return int(level)
    return level


class MissingDependencyError(ImportError):
    """
    Missing optional dependency.
    This exception is raised only when a module depends on a dependency that has not been installed.
    """

    def __init__(self, package: str, install_package: str | None = None) -> None:
        err_msg = (
            f'Package {package!r} is not installed but required. You can install it by running '  # noqa: WPS237
            f'pip install {install_package or package}'
        )
        super().__init__(err_msg)


def resolve_handlers(handlers: list[Any]) -> list[Any]:
    """
    Converts list of string of handlers to the list of object of respective handlers.

    The logging.config module uses three internal data structures to hold items
    that may need to be converted to a handler or other object.
    ConvertingList, ConvertingTuple, and ConvertingDict.

    These three objects provide interfaces to get converted items using the __getitem__ methods.

    So indexing the list performs the evaluation of the objects

    Due to missing typing in 'typeshed' we cannot type this as ConvertingList for now.
    :param handlers: An instance if 'ConvertingList'
    :return: A list of resolved handlers
    """
    return list(handlers)



import os
from collections.abc import Generator
from types import MappingProxyType

import pytest

from tests.helpers_for_tests import cleanup_logging_impl

# Using MappingProxyType so it's frozen
TEMP_ENV_VARS = MappingProxyType({
    'ENV': 'env-dev-test',
    'APP_NAME': 'system-test',
    'POD_NAME': 'pod-test',
})


@pytest.fixture(scope='session', autouse=True)
def _setup_temp_env_variables() -> Generator[None, None, None]:
    old_env = dict(os.environ)
    os.environ.update(TEMP_ENV_VARS)
    yield
    os.environ.clear()
    os.environ.update(old_env)


@pytest.fixture(autouse=True)
def _cleanup_logging() -> Generator[None, None, None]:
    with cleanup_logging_impl():
        yield



import json
import logging
from types import ModuleType

import picologging
import pytest

import modern_pylogging
from modern_pylogging.json_formatter import JsonFormatterLogging, JsonFormatterPicologging


@pytest.mark.parametrize('json_module', ['json', 'orjson', None])
@pytest.mark.parametrize('logging_module', [logging, picologging])
def test_formatter(json_module: str, logging_module: ModuleType) -> None:
    if logging_module.__name__ == 'picologging':
        formatter = JsonFormatterPicologging(json_dumps_module=json_module)  # type:ignore[arg-type]
    else:
        formatter = JsonFormatterLogging(json_dumps_module=json_module)  # type:ignore[arg-type,assignment]

    test_record = logging_module.LogRecord(
        'name', logging_module.INFO, 'path', lineno=1, msg='Log message %s', args=('123',), exc_info=None
    )
    expected_json = {
        'env': 'env-dev-test',
        'inst': 'pod-test',
        'system': 'system-test',
        'level': 'INFO',
        'message': 'Log message 123',
        'params': {'call_filepath': 'path:1', 'logger_name': 'name'},
    }
    decoded_json = json.loads(formatter.format(test_record))
    decoded_json.pop('timestamp')
    assert decoded_json == expected_json


@pytest.mark.parametrize('logging_module', [logging, picologging])
def test_extra_param_can_be_ignored(logging_module: ModuleType) -> None:
    """
    Test that JsonFormatter ignores parameters passed via extra={} if capture_extra_fields is disabled
    """
    formatter = JsonFormatterPicologging() if logging_module.__name__ == 'picologging' else JsonFormatterLogging()
    test_record = logging_module.LogRecord(
        'name', logging_module.INFO, 'path', lineno=1, msg='Log message %s', args=('123',), exc_info=None
    )
    # when we log like logger.info('bla bla', extra={'test_param': 'test_value'})
    # logging just adds  those parameters to LogRecord's __dict__
    test_record.__dict__['test_param'] = 'test_value'
    test_record.__dict__['request.id'] = 'some request id'

    decoded_json = json.loads(formatter.format(test_record))

    assert 'test_param' not in decoded_json
    assert 'request' not in decoded_json


@pytest.mark.parametrize('logging_module', [logging, picologging])
@pytest.mark.asyncio
async def test_extra_param_override_contextvars_params(logging_module: ModuleType) -> None:  # noqa: RUF029
    """
    Test that if capture_extra_fields is enabled them extra={} params have priority over
    modern_pylogging.update_log_extra fields stored in contextvars
    """
    modern_pylogging.update_log_extra({'test_param': 'test param from ctx_var'})
    if logging_module.__name__ == 'picologging':
        formatter = JsonFormatterPicologging(capture_extra_fields=True)
    else:
        formatter = JsonFormatterLogging(capture_extra_fields=True)  # type:ignore[assignment]
    test_record = logging_module.LogRecord(
        'name', logging_module.INFO, 'path', lineno=1, msg='Log message %s', args=('123',), exc_info=None
    )
    # when we log like logger.info('bla bla', extra={'test_param': 'test_value'})
    # logging just adds  those parameters to LogRecord's __dict__
    test_record.__dict__['test_param'] = 'test_value from extra='

    decoded_json = json.loads(formatter.format(test_record))

    assert decoded_json['test_param'] == 'test_value from extra='



"""
I grabbed some types straight from https://github.com/django/asgiref/blob/main/asgiref/typing.py


Copyright (c) Django Software Foundation and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import sys
from collections.abc import Callable, Iterable
from typing import Any, Literal, Protocol, TypedDict

if sys.version_info >= (3, 11):
    from typing import NotRequired, TypeAlias
else:
    from typing import TypeAlias  # noqa: WPS474

    from typing_extensions import NotRequired

Method: TypeAlias = Literal['GET', 'POST', 'DELETE', 'PATCH', 'PUT', 'HEAD', 'TRACE', 'OPTIONS']


class Logger(Protocol):  # noqa: WPS214
    """Logger protocol."""

    def debug(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'DEBUG' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def info(self, event: str, *args: Any, **kwargs: Any) -> Any:  # noqa: WPS110
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warning(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARNING' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warn(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARN' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def error(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'ERROR' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def fatal(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'FATAL' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def exception(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Log a message with level 'ERROR' on this logger. The arguments are interpreted as for debug(). Exception info
        is added to the logging message.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def critical(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def setLevel(self, level: int) -> None:  # noqa: N802
        """Set the log level

        Args:
            level: Log level to set as an integer

        Returns:
            None
        """


class ASGIVersions(TypedDict):
    spec_version: str
    version: Literal['2.0', '3.0']


class HTTPScope(TypedDict):
    type: Literal['http']
    asgi: ASGIVersions
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str, int] | None
    server: tuple[str, int | None]
    state: NotRequired[dict[str, Any]]
    extensions: dict[str, dict[object, object]] | None


class WebSocketScope(TypedDict):
    type: Literal['websocket']
    asgi: ASGIVersions
    http_version: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str, int] | None
    server: tuple[str, int | None] | None
    subprotocols: Iterable[str]
    state: NotRequired[dict[str, Any]]
    extensions: dict[str, dict[object, object]] | None


WWWScope: TypeAlias = HTTPScope | WebSocketScope
ExceptionLoggingHandler: TypeAlias = Callable[[Logger, WWWScope, list[str]], None]
GetLogger: TypeAlias = Callable[..., Logger]



import os

from modern_pylogging.helper_utils import get_logging_level
from modern_pylogging.json_formatter import get_logging_defaults


def test_get_logging_data_without_optional_envs() -> None:
    assert get_logging_defaults() == {
        'env': 'env-dev-test',
        'system': 'system-test',
        'inst': 'pod-test',
    }


def test_get_logging_level_str() -> None:
    os.environ['LOGGING_LEVEL'] = 'DEBUG'
    assert get_logging_level() == 'DEBUG'
    os.environ['LOGGING_LEVEL'] = '1'
    assert get_logging_level() == 1
    os.environ.pop('LOGGING_LEVEL', None)



import copy
import sys
import warnings
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, cast

from modern_pylogging import import_checker
from modern_pylogging.helper_types import GetLogger
from modern_pylogging.helper_utils import MissingDependencyError, get_logging_level

if TYPE_CHECKING:
    from collections.abc import Callable

    from modern_pylogging.helper_types import Logger

default_handlers: dict[str, dict[str, Any]] = {
    'console': {
        'class': 'logging.StreamHandler',
        'formatter': 'json_fmt',
        'stream': 'ext://sys.stdout',
    },
    'queue_handler': {
        'class': 'modern_pylogging.logging_handlers.QueueListenerHandler',
        'formatter': 'json_fmt',
        'respect_handler_level': True,
    },
}
if sys.version_info >= (3, 12, 0):  # pragma: no cover
    default_handlers['queue_handler'].update({
        'class': 'modern_pylogging.logging_handlers.QueueHandlerContextVarsHappyPy312',
        'queue': {
            '()': 'queue.Queue',
            'maxsize': -1,
        },
        'listener': 'modern_pylogging.logging_handlers.LoggingQueueListener',
        'handlers': ['console'],
        'respect_handler_level': True,
    })
    # do not format twice, the console handler will do the job
    default_handlers['queue_handler'].pop('formatter')

# There's no 'handlers' parameter in queue_handler
# read about it in `picologging_handlers.QueueListenerHandler`
default_picologging_handlers: dict[str, dict[str, Any]] = {
    'console': {
        # by default the class picologging.StreamHandler does not support 'stream' parameter.
        # It puts all output straight in sys.stderr
        # So I needed a little wrapper which can get 'stream' parameter and initialize StreamHandler
        'class': 'modern_pylogging.picologging_handlers.PicoStreamHandlerToStdout',
        'formatter': 'json_fmt',
        'stream': 'ext://sys.stdout',
    },
    'queue_handler': {'class': 'modern_pylogging.picologging_handlers.QueueListenerHandler', 'formatter': 'json_fmt'},
}


def _get_default_formatters() -> dict[str, dict[str, Any]]:
    return {
        'standard': {
            '()': 'modern_pylogging.console_formatter.ConsoleFormatterLogging',
            'format': '%(asctime)s | %(levelname)s | %(name)s:%(module)s:%(lineno)s | %(message)s | %(extra_data_str)s',
        },
        'json_fmt': {
            '()': 'modern_pylogging.json_formatter.JsonFormatterLogging',
            'json_dumps_module': 'json',
        },
    }


def _get_default_logging_module() -> Literal['logging', 'picologging']:
    if import_checker.is_picologging_installed:
        return 'picologging'
    return 'logging'


def _get_default_json_dumps_module() -> Literal['json', 'orjson']:
    if import_checker.is_orjson_installed:
        return 'orjson'
    return 'json'


def _get_default_handlers(logging_module: str) -> dict[str, dict[str, Any]]:
    if logging_module == 'picologging':
        return default_picologging_handlers
    return default_handlers


@dataclass
class LoggingConfig:
    """
    Configuration class for setting default python logging system.

    Attributes:
        logging_module (Literal['logging', 'picologging']):
            Specifies the backend logging module to use. 'logging' is the standard library,
            while 'picologging' is a faster, drop-in replacement.
            'picologging' will be used by default if installed.

        json_dumps_module (Literal['json', 'orjson']):
            Module used to serialize log records to JSON format. 'orjson' provides better
            performance and type handling. 'orjson' will be used by default if installed.

        version (Literal[1]):
            Schema version for the logging configuration.
            The only valid value at present is 1

        disable_existing_loggers (bool):
            Whether to disable loggers that were previously configured before applying this config.

        filters (dict[str, dict[str, Any]] | None):
            A dict in which each key is a filter id and each value is a dict describing how to configure the
            corresponding Filter_ instance.
            https://docs.python.org/3/library/logging.html#filter-objects.

        formatters (dict[str, dict[str, Any]] | None):
            A dict in which each key is a formatter and each value is a dict describing how to configure the
            corresponding Formatter_ instance. A 'standard' and 'json_fmt' formatter is provided.
            standard - write everything as strings, good for development.
            json_fmt - write output as json, good for production.
            https://docs.python.org/3/library/logging.html#formatter-objects

        handlers (dict[str, dict[str, Any]] | None):
            A dict in which each key is a handler id and each value is a dict describing how to configure the
            corresponding Handler_ instance. Two handlers are provided, 'console' and 'queue_handler'.
            https://docs.python.org/3/library/logging.html#handler-objects

        loggers (dict[str, dict[str, Any]] | None):
            A dict in which each key is a logger name and each value is a dict describing how to configure the
            corresponding Logger_ instance.
            https://docs.python.org/3/library/logging.html#logger-objects

        root (dict[str, dict[str, Any] | list[Any] | str] | None):
            This will be the configuration for the root logger.
            Processing of the configuration will be as for any logger,
            except that the propagate setting will not be applicable.

        override_formatters (dict[HandlerName, FormatterName] | None):
            Sometimes you want just use default settings but change formatters
            (e.g. to write to console as text instead of json)
            You could provide a dict like {'console': 'default', 'queue_handler': 'default'}

        level (int | str | None):
            Optional default logging level to be applied globally (e.g., 'DEBUG', 10).

        capture_extra_fields (bool):
            Whether to capture additional fields from log records' `extra` parameter
            into the structured output (especially relevant for JSON loggers).
            This does not work for 'picologging'.
            It's where you use like this - logger.info('abc', extra={'key': 'value'})
    """

    logging_module: Literal['logging', 'picologging'] = field(default_factory=_get_default_logging_module)
    json_dumps_module: Literal['json', 'orjson'] = field(default_factory=_get_default_json_dumps_module)
    version: Literal[1] = 1
    disable_existing_loggers: bool = False
    filters: dict[str, dict[str, Any]] = field(default_factory=dict)
    formatters: dict[str, dict[str, Any]] = field(default_factory=_get_default_formatters)
    handlers: dict[str, dict[str, Any]] = field(default_factory=dict)
    loggers: dict[str, dict[str, Any]] = field(default_factory=dict)
    root: dict[str, dict[str, Any] | list[Any] | str] = field(default_factory=lambda: {'handlers': ['queue_handler']})
    override_formatters: dict[str, str] = field(default_factory=dict)
    level: int | str = field(default_factory=get_logging_level)
    capture_extra_fields: bool = False

    def __post_init__(self) -> None:  # noqa: C901, WPS231
        if self.logging_module == 'picologging' and import_checker.is_picologging_installed is False:
            raise MissingDependencyError('picologging')
        if self.json_dumps_module == 'orjson' and import_checker.is_orjson_installed is False:
            raise MissingDependencyError('orjson')
        if 'standard' not in self.formatters:
            self.formatters['standard'] = _get_default_formatters()['standard']
        if 'json_fmt' not in self.formatters:
            self.formatters['json_fmt'] = _get_default_formatters()['json_fmt']
        if 'console' not in self.handlers:
            self.handlers['console'] = _get_default_handlers(self.logging_module)['console']
        if 'queue_handler' not in self.handlers:
            self.handlers['queue_handler'] = _get_default_handlers(self.logging_module)['queue_handler']
        if self.logging_module == 'picologging':
            if self.formatters['json_fmt'].get('()', '') == 'modern_pylogging.json_formatter.JsonFormatterLogging':
                self.formatters['json_fmt']['()'] = 'modern_pylogging.json_formatter.JsonFormatterPicologging'
            if (
                self.formatters['standard'].get('()', '')
                == 'modern_pylogging.console_formatter.ConsoleFormatterLogging'
            ):
                self.formatters['standard']['()'] = 'modern_pylogging.console_formatter.ConsoleFormatterPicologging'

    def configure(self) -> GetLogger:
        config_dict = self._prepare_config_dict()
        if self.logging_module == 'picologging':
            try:
                from picologging import config, getLogger
            except ImportError as exc:
                raise MissingDependencyError('picologging') from exc
        else:
            from logging import config, getLogger  # type:ignore[assignment,no-redef]
        config.dictConfig(config_dict)
        return cast('Callable[[str], Logger]', getLogger)

    def replace_formatters(self, override_formatters: dict[str, str], source: dict[str, Any]) -> dict[str, Any]:
        if override_formatters:
            source = copy.deepcopy(source)
            # If we simply change the `source` variable, then new parameters
            # will remain in the 'default_handlers' dictionary at the module level.
            # This is generally not an issue because configuring logging multiple times
            # in a single application is unnecessary.
            #
            # However, this becomes problematic during tests
            # If I use `override_formatters` in one test, those values are automatically carried over to other tests.
            # Interestingly, this behavior only affects tests located in the same .py file.
            # It's possible that pytest loads modules from scratch for tests in different files.
            # For example, the first test sets the standard format globally using `override_formatters`
            # and verifies the console output.
            # In the same file, another tests checks if the JsonFormatter correctly calls
            # the function to dump data in JSON.
            # Since `override_formatters` is not used during the second test
            # (as json_fmt is the default formatter everythere),
            # the previous formatter settings from the first test remains, causing the second test to fail.
            # The occurrence of this failure is unpredictable
            # depending on the execution order of the tests and whether they are in the same file.
            # Debugging this issue was quite time-consumig for me.
            for handler_name, formatter_name in override_formatters.items():
                if source['handlers'][handler_name].get('formatter') is not None:
                    # if 'formatter' in values['handlers'][handler_name]:
                    source['handlers'][handler_name]['formatter'] = formatter_name
        return source

    def _prepare_config_dict(self) -> dict[str, Any]:
        config_base = {
            'version': self.version,
            'disable_existing_loggers': self.disable_existing_loggers,
            'filters': self.filters,
            'formatters': self.formatters,
            'handlers': self.handlers,
            'loggers': self.loggers,
            'root': self.root,
        }
        config_dict = {name: value for name, value in config_base.items() if value}  # noqa: WPS110
        config_dict['formatters']['json_fmt']['json_dumps_module'] = self.json_dumps_module  # type:ignore[index] # noqa: WPS204
        config_dict['formatters']['json_fmt']['capture_extra_fields'] = self.capture_extra_fields  # type:ignore[index]
        config_dict['root']['level'] = self.level  # type:ignore[index]
        if self.logging_module == 'picologging' and 'datefmt' not in config_dict['formatters']['standard']:  # type:ignore[index]
            # There is a problem in picologging with time formatting with single formatter
            # https://github.com/microsoft/picologging/issues/203
            # https://github.com/microsoft/picologging/pull/208
            config_dict['formatters']['standard']['datefmt'] = '%F %T'  # type: ignore[index]

        config_dict = self.replace_formatters(self.override_formatters, config_dict)
        _python_3_12_4_fix(config_dict)
        if self.logging_module == 'picologging' and self.capture_extra_fields:
            _picologging_and_capture_extra_fields_warning()
            config_dict['formatters']['json_fmt']['capture_extra_fields'] = False
        return config_dict


def _python_3_12_4_fix(config_dict: dict[str, Any]) -> None:  # noqa: WPS114
    if sys.version_info[:3] == (3, 12, 4):
        if 'queue' in config_dict['handlers']['queue_handler']:
            warnings.warn(
                'Python 3.12.4 has a regression bug '
                'and it fails to configure `queue` parameter for QueueHandler '
                'https://github.com/python/cpython/issues/119818',
                stacklevel=2,
            )
            config_dict['handlers']['queue_handler'].pop('queue')


def _picologging_and_capture_extra_fields_warning() -> None:
    warnings.warn(
        'log `extra` param does not work with picologging '
        'so parameter `capture_extra_fields` is disabled '
        'https://github.com/microsoft/picologging/issues/195',
        stacklevel=2,
    )



import atexit
import logging
import sys
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import Any, cast

from modern_pylogging.contextvars_helpers import get_log_extra
from modern_pylogging.helper_utils import resolve_handlers


class LoggingQueueListener(QueueListener):
    """Custom `QueueListener` that starts and stops the listening thread."""

    def __init__(
        self,
        queue: Queue[logging.LogRecord],
        *handlers: logging.Handler,
        respect_handler_level: bool = False,
    ) -> None:
        """
        Initialize a `LoggingQueueListener` instance.
        :param queue: The queue to send messages to
        :param handlers: A list of handlers whichj will handle entries placed on the queue
        :param respect_handler_level: If `respect_handler` is `True`
        a handler's level is respected (compared with the level for the message)
        when deciding whether to pass message to that handler.
        """
        super().__init__(queue, *handlers, respect_handler_level=respect_handler_level)
        self.start()
        atexit.register(self.stop)


class QueueListenerHandler(QueueHandler):
    """
    Configure queue listener and handler to support non-blocking logging configuration.

    !caution!
    This handler DOES NOT WORK with Python >= 3.12 and `logging.config.dictConfig`.
    Please use `QueueHandlerContextVarsHappyPy312`
    """

    def __init__(self, handlers: list[Any] | None = None, respect_handler_level: bool = False) -> None:  # noqa: FBT001, FBT002
        super().__init__(Queue(-1))
        handlers = resolve_handlers(handlers) if handlers else [logging.StreamHandler(stream=sys.stdout)]
        self.listener = LoggingQueueListener(
            self.queue,  # type:ignore[arg-type]
            *handlers,
            respect_handler_level=respect_handler_level,
        )


class QueueHandlerContextVarsHappyPy312(QueueHandler):
    """
    In python up to version 3.12 JsonFormatter was called in the same thread as logging
    and placed already formatted string directly into the queue.

    However, in python 3.12 this behavior was changed.
    Now, in the main thread, we simply enqueue LogRecords,
    and the queue processing is handled by QueueListener in a separate thread.
    Since ContextVars are not shared between threads, any data that was set using `update_log_extra` will be lost.
    To prevent this from happening, I created this handler and modified the JsonFormatter.
    """

    def prepare(self, record: logging.LogRecord) -> logging.LogRecord:
        record = super().prepare(record)
        record.binded_ctx_extra = get_log_extra()
        return cast('logging.LogRecord', record)



import atexit
import contextlib
import logging
import sys
from collections.abc import Generator
from typing import Any

import picologging
from _pytest.logging import LogCaptureHandler, _LiveLoggingNullHandler

if sys.version_info >= (3, 12):
    getHandlerByName = logging.getHandlerByName  # noqa: N816
else:

    def getHandlerByName(name: str) -> Any:  # noqa: N802
        return logging._handlers.get(name)  # type:ignore[attr-defined]


@contextlib.contextmanager
def cleanup_logging_impl() -> Generator[None, None, None]:
    # Reset logging settings (`logging` module)
    # - delete non-standart loggers created by logging.getLogger in tests
    # - remove all handlers except pytest's ones
    # - stop queue_handler listener
    # so tests could configure logging without affecting each other
    std_root_logger: logging.Logger = logging.getLogger()
    name_of_loggers_exist_on_start = set(std_root_logger.manager.loggerDict)
    for std_handler in std_root_logger.handlers:
        # Do not interfere with pytest handler config
        if not isinstance(std_handler, (_LiveLoggingNullHandler, LogCaptureHandler)):
            std_root_logger.removeHandler(std_handler)

    # reset root logger (`picologging` module)
    pico_root_logger: picologging.Logger = picologging.getLogger()
    for pico_handler in pico_root_logger.handlers:
        pico_root_logger.removeHandler(pico_handler)

    yield

    # Stop queue_handler's listener (mandatory for the 'logging' module with Python 3.12,
    # else the test suite would hand on at the end of the tests and some tests would fail)
    queue_listener_handler = getHandlerByName('queue_handler')
    if queue_listener_handler and hasattr(queue_listener_handler, 'listener'):
        atexit.unregister(queue_listener_handler.listener.stop)
        queue_listener_handler.listener.stop()
        queue_listener_handler.close()
        del queue_listener_handler  # noqa: WPS420

    name_of_loggers_exist_on_end = set(std_root_logger.manager.loggerDict)
    for name in name_of_loggers_exist_on_end - name_of_loggers_exist_on_start:
        std_root_logger.manager.loggerDict.pop(name)



import json
import logging
import sys
import time
from importlib.util import find_spec
from queue import Queue
from types import ModuleType
from typing import Any, cast

import picologging
import pytest
from _pytest.capture import CaptureFixture
from pytest_mock import MockerFixture

from modern_pylogging import logging_handlers, picologging_handlers
from modern_pylogging.config_api import (
    LoggingConfig,
    _get_default_handlers,
    _get_default_json_dumps_module,
    _get_default_logging_module,
    default_handlers,
    default_picologging_handlers,
)
from modern_pylogging.json_formatter import JsonFormatterLogging, JsonFormatterPicologging


def test_get_default_handlers() -> None:
    assert _get_default_handlers(logging_module='logging') == default_handlers
    assert _get_default_handlers(logging_module='picologging') == default_picologging_handlers


def test_get_default_logging_module(mocker: MockerFixture) -> None:
    assert find_spec('picologging')  # should be installed in the test env, simply checking
    assert _get_default_logging_module() == 'picologging'
    mock = mocker.patch('modern_pylogging.config_api.import_checker')
    mock.is_picologging_installed = False
    assert _get_default_logging_module() == 'logging'


def test_get_default_json_module(mocker: MockerFixture) -> None:
    assert find_spec('orjson')  # should be installed in the test env, simply checking
    assert _get_default_json_dumps_module() == 'orjson'
    mock = mocker.patch('modern_pylogging.config_api.import_checker')
    mock.is_orjson_installed = False
    assert _get_default_json_dumps_module() == 'json'


@pytest.mark.parametrize(
    ('logging_module', 'dict_config_callable', 'expected_called'),
    [
        ('logging', 'logging.config.dictConfig', True),
        ('logging', 'picologging.config.dictConfig', False),
        ('picologging', 'picologging.config.dictConfig', True),
        ('picologging', 'logging.config.dictConfig', False),
    ],
)
def test_correct_dict_config_called(
    logging_module: str,
    dict_config_callable: str,
    expected_called: bool,  # noqa: FBT001
    mocker: MockerFixture,
) -> None:
    dict_config_mock = mocker.patch(dict_config_callable)
    LoggingConfig(logging_module=logging_module).configure()  # type:ignore[arg-type]
    assert dict_config_mock.called is expected_called


@pytest.mark.parametrize(
    ('picologging_installed', 'expected_default_handlers'),
    [(True, default_picologging_handlers), (False, default_handlers)],
)
def test_correct_default_handlers_set(
    picologging_installed: bool,  # noqa: FBT001
    expected_default_handlers: dict[str, dict[str, Any]],
    mocker: MockerFixture,
) -> None:
    mock = mocker.patch('modern_pylogging.config_api.import_checker')
    mock.is_picologging_installed = picologging_installed
    log_config = LoggingConfig()._prepare_config_dict()
    assert log_config['handlers'] == expected_default_handlers


@pytest.mark.parametrize(('orjson_exists', 'json_module_name'), [(True, 'orjson'), (False, 'json')])
def test_correct_json_module_set(
    orjson_exists: bool,  # noqa: FBT001
    json_module_name: str,
    mocker: MockerFixture,
) -> None:
    mock = mocker.patch('modern_pylogging.config_api.import_checker')
    mock.is_orjson_installed = orjson_exists
    log_config = LoggingConfig()._prepare_config_dict()
    assert log_config['formatters']['json_fmt']['json_dumps_module'] == json_module_name


@pytest.mark.parametrize(
    ('logging_module', 'expected_handlers'),
    [('picologging', default_picologging_handlers), ('logging', default_handlers)],
)
def test_correct_handlers_set_by_param(
    logging_module: str,
    expected_handlers: dict[str, dict[str, Any]],
) -> None:
    log_config = LoggingConfig(logging_module=logging_module)._prepare_config_dict()  # type:ignore[arg-type]
    assert log_config['handlers'] == expected_handlers


@pytest.mark.parametrize(('json_module_name', 'expected_module_name'), [('orjson', 'orjson'), ('json', 'json')])
def test_correct_json_module_set_by_param(
    json_module_name: str,
    expected_module_name: dict[str, dict[str, Any]],
) -> None:
    log_config = LoggingConfig(json_dumps_module=json_module_name)._prepare_config_dict()  # type: ignore[arg-type]
    assert log_config['formatters']['json_fmt']['json_dumps_module'] == expected_module_name


def test_capture_extra_fields_cannot_be_used_with_picologging() -> None:  # noqa: WPS118
    log_config = LoggingConfig(logging_module='picologging', capture_extra_fields=True)._prepare_config_dict()
    assert log_config['formatters']['json_fmt']['capture_extra_fields'] is False


@pytest.mark.parametrize(
    ('logging_module', 'expected_handler_class', 'expected_listener_class'),
    [
        (
            logging,
            logging_handlers.QueueHandlerContextVarsHappyPy312
            if sys.version_info >= (3, 12, 0)
            else logging_handlers.QueueListenerHandler,
            logging_handlers.LoggingQueueListener,
        ),
        (
            picologging,
            picologging_handlers.QueueListenerHandler,
            picologging.handlers.QueueListener,
        ),
    ],
)
def test_default_queue_listener_handler(
    logging_module: ModuleType,
    expected_handler_class: Any,
    expected_listener_class: Any,
    capsys: CaptureFixture[str],
) -> None:
    get_logger = LoggingConfig(
        logging_module=logging_module.__name__,  # type:ignore[arg-type]
        formatters={'standard': {'format': '%(levelname)s :: %(name)s :: %(message)s'}},
        loggers={
            'test_logger': {
                'level': 'INFO',
                'handlers': ['queue_handler'],
                'propagate': False,
            },
        },
        override_formatters={
            'queue_handler': 'standard',
            'console': 'standard',
        },
    ).configure()

    logger = get_logger('test_logger')
    assert isinstance(logger, logging_module.Logger)

    handler = logger.handlers[0]
    assert isinstance(handler, expected_handler_class)
    assert isinstance(handler.queue, Queue)

    assert isinstance(handler.listener, expected_listener_class)
    assert isinstance(handler.listener.handlers[0], logging_module.StreamHandler)

    logger.info('Testing now!')
    assert_log(capsys, handler.queue, expected='INFO :: test_logger :: Testing now!', count=1)

    var = 'test_var'
    logger.info('%s', var)
    assert_log(capsys, handler.queue, expected='INFO :: test_logger :: test_var', count=1)


@pytest.mark.parametrize(
    ('logging_module', 'expected_handler_class'),
    [
        (
            logging,
            logging_handlers.QueueHandlerContextVarsHappyPy312
            if sys.version_info >= (3, 12, 0)
            else logging_handlers.QueueListenerHandler,
        ),
        (picologging, picologging_handlers.QueueListenerHandler),
    ],
)
def test_default_logger_and_handler(logging_module: ModuleType, expected_handler_class: Any) -> None:
    get_logger = LoggingConfig(logging_module=logging_module.__name__).configure()  # type:ignore[arg-type]

    root_logger = get_logger()

    assert isinstance(root_logger, logging_module.Logger)
    assert root_logger.name == 'root'
    assert isinstance(root_logger.handlers[0], expected_handler_class)
    assert root_logger.handlers[0].name == 'queue_handler'


@pytest.mark.parametrize('logging_module', [logging, picologging, None])
def test_default_handlers_formatters(logging_module: ModuleType | None) -> None:
    if logging_module is None:
        config = LoggingConfig(
            formatters={},
            handlers={},
            loggers={},
        )._prepare_config_dict()
        expected_default_handlers = _get_default_handlers(_get_default_logging_module())
    else:
        config = LoggingConfig(
            logging_module=logging_module.__name__,  # type:ignore[arg-type]
            formatters={},
            handlers={},
            loggers={},
        )._prepare_config_dict()
        expected_default_handlers = _get_default_handlers(logging_module.__name__)

    assert config['formatters']['json_fmt']
    assert len(config['formatters']) == 2
    assert config['handlers']['queue_handler'] == expected_default_handlers['queue_handler']
    assert config['handlers']['console'] == expected_default_handlers['console']
    assert len(config['handlers']) == 2


@pytest.mark.parametrize(
    ('logging_module', 'expected_handler_class'),
    [
        (
            logging,
            logging_handlers.QueueHandlerContextVarsHappyPy312
            if sys.version_info >= (3, 12, 0)
            else logging_handlers.QueueListenerHandler,
        ),
        (picologging, picologging_handlers.QueueListenerHandler),
    ],
)
def test_customizing_handler(
    logging_module: ModuleType,
    expected_handler_class: Any,
    capsys: CaptureFixture[str],
) -> None:
    log_format = '%(levelname)s :: %(name)s :: %(message)s'
    # picologging seems to be broken, cannot make it log on stdout?
    # https://github.com/microsoft/picologging/issues/205
    if logging_module == picologging:
        console_stdout_class = 'modern_pylogging.picologging_handlers.PicoStreamHandlerToStdout'
    else:
        console_stdout_class = f'{logging_module.__name__}.StreamHandler'
    get_logger = LoggingConfig(
        logging_module=logging_module.__name__,  # type:ignore[arg-type]
        formatters={'standard': {'format': log_format}},
        handlers={
            'console_stdout': {
                'class': console_stdout_class,
                'stream': 'ext://sys.stdout',
                'level': 'DEBUG',
                'formatter': 'standard',
            }
        },
        loggers={
            'test_logger': {
                'level': 'DEBUG',
                'handlers': ['console_stdout'],
                'propagate': False,
            }
        },
    ).configure()

    root_logger = get_logger()
    assert isinstance(root_logger, logging_module.Logger)
    assert root_logger.level == logging_module.INFO

    root_handler = root_logger.handlers[0]
    assert root_handler.name == 'queue_handler'
    assert isinstance(root_handler, expected_handler_class)

    if isinstance(root_handler, logging_handlers.QueueHandlerContextVarsHappyPy312):
        root_formatter = root_handler.listener.handlers[0].formatter  # type:ignore[union-attr,unused-ignore,attr-defined]
    else:
        root_formatter = root_handler.formatter
    assert isinstance(root_formatter, (JsonFormatterLogging, JsonFormatterPicologging))

    t_logger = get_logger('test_logger')
    assert isinstance(t_logger, logging_module.Logger)
    assert t_logger.level == logging_module.DEBUG
    assert len(t_logger.handlers) == 1
    if logging_module == picologging:
        assert isinstance(t_logger.handlers[0], picologging_handlers.PicoStreamHandlerToStdout)
    else:
        assert isinstance(t_logger.handlers[0], logging_module.StreamHandler)

    assert t_logger.handlers[0].name == 'console_stdout'
    assert t_logger.handlers[0].formatter._fmt == log_format

    t_logger.info("Hello from '%s'", logging_module.__name__)
    log_output = capsys.readouterr().out.strip()
    assert log_output == f"INFO :: {t_logger.name} :: Hello from '{logging_module.__name__}'"


@pytest.mark.parametrize('logging_module', [logging, picologging])
def test_excluded_fields(logging_module: ModuleType, mocker: MockerFixture) -> None:
    # according to https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
    allowed_fields = {
        'version',
        'formatters',
        'filters',
        'handlers',
        'loggers',
        'root',
        'incremental',
        'disable_existing_loggers',
        # 'level',
        # plus our custom
        # 'json_dumps_module'
    }
    mocked = mocker.patch(f'{logging_module.__name__}.config.dictConfig')
    LoggingConfig(logging_module=logging_module.__name__).configure()  # type:ignore[arg-type]
    assert mocked.called
    for key in mocked.call_args.args[0]:
        assert key in allowed_fields


@pytest.mark.parametrize(
    ('json_module', 'expected_json_func'),
    [
        ('json', 'modern_pylogging.json_helper.json_dumps'),
        ('orjson', 'modern_pylogging.orjson_helper.json_dumps'),
    ],
)
def test_json_formatter_uses_json_func(
    json_module: str,
    expected_json_func: str,
    mocker: MockerFixture,
) -> None:
    mocked = mocker.patch(expected_json_func)
    mocked.return_value = '{"msg": "ok"}'
    get_logger = LoggingConfig(
        logging_module='logging',
        json_dumps_module=json_module,  # type:ignore[arg-type]
        root={'handlers': ['console']},
    ).configure()
    logger = get_logger('test_logger')
    logger.info('Hello from %s', json_module)

    assert mocked.called
    assert mocked.call_args.args[0]['message'] == f'Hello from {json_module}'


def test_default_queue_listener_handler_json_fmt_extra(capsys: CaptureFixture[str]) -> None:  # noqa: WPS118
    """
    Test that we can do logger.info('msg', extra={'hehe': 'pepe'})
    and if configure logging with `capture_extra_fields` option
    these extra parameters will be passed to the logging formatter
    and will be present in json log (but only for logrecord in which they were provided)
    """
    get_logger = LoggingConfig(
        capture_extra_fields=True,
        logging_module='logging',
        loggers={'test_logger': {'level': 'INFO', 'handlers': ['queue_handler'], 'propagate': False}},
        # propagate=False will save us from doubled logs
        override_formatters={'queue_handler': 'json_fmt'},
    ).configure()
    logger = get_logger('test_logger')
    handler = logger.handlers[0]  # type:ignore[attr-defined]
    logger.info(
        'Testing one!',
        extra={'simple': 'some text', 'nested.id': 'n_id', 'nested.msg': 'n_msg', 'params.additional': 'something'},
    )
    log_entry = capture_json_log(capsys, handler.queue)

    assert log_entry['simple'] == 'some text'
    assert log_entry['nested'] == {
        'id': 'n_id',
        'msg': 'n_msg',
    }
    assert log_entry['params']['additional'] == 'something'
    assert log_entry['params']['logger_name'] == 'test_logger'  # update_log_extra does not delete existing keys

    # logging with extra={} does not affect the following log records
    logger.info('Testing two!')
    log_entry = capture_json_log(capsys, handler.queue)
    assert 'simple' not in log_entry
    assert 'nested' not in log_entry
    assert 'additional' not in log_entry['params']


def wait_log_queue(queue: Any, sleep_time: float = 0.1, max_retries: int = 5) -> None:
    retry = 0
    while queue.qsize() > 0 and retry < max_retries:
        retry += 1
        time.sleep(sleep_time)


def capture_json_log(capsys: CaptureFixture[str], queue: Any) -> dict[str, Any]:
    wait_log_queue(queue)
    log_output = (
        capsys.readouterr().out.strip()
    )  # we log to stdout by default instead of stderr (even with picologging)
    return cast('dict[str, Any]', json.loads(log_output))


def assert_log(capsys: CaptureFixture[str], queue: Any, expected: str, count: int | None = None) -> None:
    wait_log_queue(queue)
    log_output = capsys.readouterr().out.strip()  # we log to stdout by default instead of stderr
    if count is not None:
        assert len(log_output.split('\n')) == count
    assert log_output == expected



"""
Functions to set context for async task and its children.
Article about it - https://habr.com/ru/companies/yandex/articles/526322/
"""

import copy
from contextvars import ContextVar
from typing import Any

_log_extra_data: ContextVar[dict[str, Any]] = ContextVar('_log_extra_data')


def set_log_extra(log_extra: dict[str, Any]) -> None:
    _log_extra_data.set(log_extra)


def get_log_extra(should_copy: bool = False) -> dict[str, Any]:  # noqa: FBT001,FBT002
    try:
        extra = _log_extra_data.get()
    except LookupError:
        # sometimes we log from background tasks, etc...
        # so contextvar wouldn't be set before
        # to avoid explicit checks I decided just to use try-except here
        extra = {}
    if should_copy:
        return copy.deepcopy(extra)
    return extra


def update_log_extra(updates: dict[str, Any]) -> None:
    # Our _log_extra contextvar contains mutable nested object (dict)
    # We need that any updates would propagate to child tasks,
    # but not to the parents - because of that we use get with copying
    current_extra = get_log_extra(should_copy=True)
    set_log_extra(current_extra | updates)



from typing import Any

from modern_pylogging import import_checker
from modern_pylogging.helper_utils import MissingDependencyError

if import_checker.is_orjson_installed:
    import orjson
else:
    raise MissingDependencyError('orjson')


def default(obj_to_serialize: Any) -> Any:
    """See https://github.com/ijl/orjson#default for information."""
    if isinstance(obj_to_serialize, set):
        return list(obj_to_serialize)
    raise TypeError


def json_dumps(obj_to_serialize: Any) -> str:
    # orjson.dumps returns bytes, to match standart json.dumps we need to decode
    return orjson.dumps(obj_to_serialize, default=default).decode()



import atexit
import sys
from queue import Queue
from typing import IO, Any

from modern_pylogging.helper_utils import MissingDependencyError, resolve_handlers

try:
    import picologging
    from picologging.handlers import QueueHandler
except ImportError as err:
    raise MissingDependencyError('picologging') from err


class PicoStreamHandlerToStdout(picologging.StreamHandler):  # type:ignore[type-arg]
    """
    By default the class picologging.StreamHandler does not support 'stream' parameter.
    It puts all output straight in sys.stderr
    So I needed a little wrapper which can get 'stream' parameter and initialize StreamHandler
    """

    def __init__(self, stream: IO[bytes | str]) -> None:  # type:ignore[type-var] # noqa: WPS612
        super().__init__(stream)


class QueueListenerHandler(QueueHandler):
    """Configure queue listener and handler to support non-blocking logging configuration."""

    def __init__(
        self,
        handlers: list[Any] | None = None,
    ) -> None:
        # handlers are ConvertingList (from logging.config
        super().__init__(Queue(-1))
        """
        In the builtin logging setup, when configured through a DictConfig
        the ConvertingList entity is passed to the handlers.
        After going through the resolve_handlers process,
        it is converted into a handler object based on the configuration.
        For example, the default configuration includes a `console` handler like this
        ['console'] and it resolves into [<StreamHandler (NOTSET)>] (python object of StreamHandler).
        However, in picologging, for some inexplicable reason,
        the ConvertingList is illogically converted into a list of strings.
        So ['console'] becomes ['console'] instead of being converted to the appropriate handler object.
        That's why in picologging config I didn't use `handlers` parameter.
        """
        handlers = resolve_handlers(handlers) if handlers else [picologging.StreamHandler(sys.stdout)]
        self.listener = picologging.handlers.QueueListener(self.queue, *handlers)
        self.listener.start()
        atexit.register(self.listener.stop)
